"""Reference elements."""

from __future__ import annotations

import typing
from abc import ABC, abstractmethod

import sympy

from .geometry import (PointType, PointTypeInput, SetOfPoints, SetOfPointsInput, parse_point_input,
                       parse_set_of_points_input)
from .symbols import AxisVariablesNotSingle, t, x

LatticeWithLines = typing.Tuple[SetOfPoints, typing.List[typing.Tuple[int, int]]]
IntLimits = typing.List[typing.Union[
    typing.Tuple[sympy.core.symbol.Symbol, sympy.core.expr.Expr, sympy.core.expr.Expr],
    typing.Tuple[sympy.core.symbol.Symbol, sympy.core.expr.Expr]]]


def _vsub(v: PointTypeInput, w: PointTypeInput) -> PointType:
    """Subtract.

    Args:
        v: A vector
        w: A vector

    Returns:
        The vector v - w
    """
    return tuple(i - j for i, j in zip(parse_point_input(v), parse_point_input(w)))


def _vadd(v: PointTypeInput, w: PointTypeInput) -> PointType:
    """Add.

    Args:
        v: A vector
        w: A vector

    Returns:
        The vector v + w
    """
    return tuple(i + j for i, j in zip(parse_point_input(v), parse_point_input(w)))


def _vdot(v: PointTypeInput, w: PointTypeInput) -> sympy.core.expr.Expr:
    """Compute the dot product.

    Args:
        v: A vector
        w: A vector

    Returns:
        The scalar v.w
    """
    out = sympy.Integer(0)
    for i, j in zip(parse_point_input(v), parse_point_input(w)):
        out += i * j
    return out


def _vcross(v_in: PointTypeInput, w_in: PointTypeInput) -> PointType:
    """Compute the cross product.

    Args:
        v: A vector
        w: A vector

    Returns:
        The vector v x w
    """
    v = parse_point_input(v_in)
    w = parse_point_input(w_in)
    assert len(v) == len(w) == 3
    return (v[1] * w[2] - v[2] * w[1], v[2] * w[0] - v[0] * w[2], v[0] * w[1] - v[1] * w[0])


def _vnorm(v_in: PointTypeInput) -> sympy.core.expr.Expr:
    """Find the norm of a vector.

    Args:
        v_in: A vector

    Returns:
        The norm of v_in
    """
    v = parse_point_input(v_in)
    return sympy.sqrt(_vdot(v, v))


def _vnormalise(v_in: PointTypeInput) -> PointType:
    """Normalise a vector.

    Args:
        v_in: A vector

    Returns:
        A unit vector pointing in the same direction as v_in
    """
    v = parse_point_input(v_in)
    n = _vnorm(v)
    return tuple(i / n for i in v)


class Reference(ABC):
    """A reference cell on which a finite element can be defined."""

    def __init__(
        self, tdim: int, name: str, origin: PointTypeInput, axes: SetOfPointsInput,
        reference_vertices: SetOfPointsInput, vertices: SetOfPointsInput,
        edges: typing.Tuple[typing.Tuple[int, int], ...],
        faces: typing.Tuple[typing.Tuple[int, ...], ...],
        volumes: typing.Tuple[typing.Tuple[int, ...], ...],
        sub_entity_types: typing.List[typing.Union[typing.List[str], str, None]],
        simplex: bool = False, tp: bool = False
    ):
        """Create a reference cell.

        Args:
            tdim: The topological dimension of the cell
            name: The name of the cell
            origin: The coordinates of the origin
            axes: Vectors representing the axes of the cell
            reference_vertices: The vertices of the default version of this cell
            vertices: The vertices of this cell
            edges: Pairs of vertex numbers that form the edges of the cell
            faces: Tuples of vertex numbers that form the faces of the cell
            volumes: Tuples of vertex numbers that form the volumes of the cell
            sub_entity_types: The cell types of each sub-entity of the cell
            simplex: Is the cell a simplex (interval/triangle/tetrahedron)?
            tp: Is the cell a tensor product (interval/quadrilateral/hexahedron)?
        """
        self.tdim = tdim
        self.origin: PointType = parse_point_input(origin)
        self.gdim = len(self.origin)
        self.name = name
        self.axes: SetOfPoints = parse_set_of_points_input(axes)
        self.reference_vertices: SetOfPoints = parse_set_of_points_input(reference_vertices)
        self.vertices: SetOfPoints = parse_set_of_points_input(vertices)
        self.edges = edges
        self.faces = faces
        self.volumes = volumes
        self.sub_entity_types = sub_entity_types
        self.simplex = simplex
        self.tp = tp
        self._inverse_map_to_self: typing.Union[PointType, None] = None
        self._map_to_self: typing.Union[PointType, None] = None

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return self.vertices

    @abstractmethod
    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        pass

    @abstractmethod
    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        pass

    @abstractmethod
    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        pass

    def make_lattice_float(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        return tuple(tuple(sympy.Float(float(i)) for i in p) for p in self.make_lattice(n))

    def make_lattice_with_lines_float(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        pts, pairs = self.make_lattice_with_lines(n)
        return tuple(tuple(sympy.Float(float(i)) for i in p) for p in pts), pairs

    def z_ordered_entities(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities of the cell in back-to-front plotting order.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [[(i, j) for i in range(self.tdim, -1, -1) for j in range(self.sub_entity_count(i))]]

    def z_ordered_entities_extra_dim(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities in back-to-front plotting order when using an extra dimension.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return self.z_ordered_entities()

    def get_point(self, reference_coords: PointType) -> typing.Tuple[sympy.core.expr.Expr, ...]:
        """Get a point in the reference from reference coordinates.

        Args:
            reference_coords: The reference coordinates

        Returns:
            A point in the cell
        """
        assert len(reference_coords) == len(self.axes)
        pt = [i for i in self.origin]
        for a, b in zip(self.axes, reference_coords):
            for i, ai in enumerate(a):
                pt[i] += ai * b
        return tuple(pt)

    @abstractmethod
    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        pass

    @abstractmethod
    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        pass

    @abstractmethod
    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        pass

    def get_map_to_self(self) -> PointType:
        """Get the map from the canonical reference to this reference.

        Returns:
            The map
        """
        if self._map_to_self is None:
            self._map_to_self = self._compute_map_to_self()
        assert self._map_to_self is not None
        return self._map_to_self

    @abstractmethod
    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        pass

    def get_inverse_map_to_self(self) -> PointType:
        """Get the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        if self._inverse_map_to_self is None:
            self._inverse_map_to_self = self._compute_inverse_map_to_self()
        assert self._inverse_map_to_self is not None
        return self._inverse_map_to_self

    @abstractmethod
    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        pass

    @abstractmethod
    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        pass

    def midpoint(self) -> PointType:
        """Calculate the midpoint.

        Returns:
            The midpoint of the cell
        """
        return tuple(sum(i) * sympy.Integer(1) / len(i) for i in zip(*self.vertices))

    def jacobian(self) -> sympy.core.expr.Expr:
        """Calculate the Jacobian.

        Returns:
            The Jacobian
        """
        from .functions import VectorFunction
        assert len(self.axes) == self.tdim
        vaxes = [VectorFunction(a) for a in self.axes]
        if self.tdim == 1:
            out = vaxes[0].norm().as_sympy()
        elif self.tdim == 2:
            out = vaxes[0].cross(vaxes[1]).norm().as_sympy()
        elif self.tdim == 3:
            out = vaxes[0].cross(vaxes[1]).dot(vaxes[2]).norm().as_sympy()
        else:
            raise ValueError(f"Unsupported tdim: {self.tdim}")
        assert isinstance(out, sympy.core.expr.Expr)
        return out

    def scaled_axes(self) -> SetOfPoints:
        """Return the unit axes of the reference.

        Returns:
            The axes
        """
        return tuple(_vnormalise(a) for a in self.axes)

    def tangent(self) -> PointType:
        """Calculate the tangent to the element.

        Returns:
            The tangent
        """
        if self.tdim == 1:
            norm = sympy.sqrt(sum(i ** 2 for i in self.axes[0]))
            return _vnormalise(tuple(i / norm for i in self.axes[0]))

        raise RuntimeError

    def normal(self) -> PointType:
        """Calculate the normal to the element.

        Returns:
            The normal
        """
        if self.tdim == 1:
            if self.gdim == 2:
                return _vnormalise((-self.axes[0][1], self.axes[0][0]))
        if self.tdim == 2:
            if self.gdim == 3:
                crossed = _vcross(self.axes[0], self.axes[1])
                assert isinstance(crossed, tuple)
                return _vnormalise(crossed)
        raise RuntimeError

    def sub_entities(
        self, dim: typing.Optional[int] = None, codim: typing.Optional[int] = None
    ) -> typing.Tuple[typing.Tuple[int, ...], ...]:
        """Get the sub-entities of a given dimension.

        Args:
            dim: The dimension of the sub-entity
            codim: The co-dimension of the sub-entity

        Returns:
            A tuple of tuples of vertex numbers
        """
        if dim is None:
            assert codim is not None
            dim = self.tdim - codim
        if dim == 0:
            return tuple((i, ) for i, _ in enumerate(self.vertices))
        if dim == 1:
            return self.edges
        if dim == 2:
            return self.faces
        if dim == 3:
            return self.volumes
        raise ValueError(f"Unsupported dimension: {dim}")

    def sub_entity_count(
        self, dim: typing.Optional[int] = None, codim: typing.Optional[int] = None
    ) -> int:
        """Get the number of sub-entities of a given dimension.

        Args:
            dim: the dimension of the sub-entity
            codim: the codimension of the sub-entity

        Returns:
            The number of sub-entities
        """
        return len(self.sub_entities(dim, codim))

    def sub_entity(self, dim: int, n: int, reference_vertices: bool = False) -> typing.Any:
        # def sub_entity(self, dim: int, n: int, reference_vertices: bool = False) -> Reference:
        """Get the sub-entity of a given dimension and number.

        Args:
            dim: the dimension of the sub-entity
            n: The sub-entity number
            reference_vertices: Should the reference vertices be used?

        Returns:
            The sub-entity
        """
        from symfem import create_reference

        entity_type = self.sub_entity_types[dim]
        if not isinstance(entity_type, str):
            assert isinstance(entity_type, list)
            entity_type = entity_type[n]

        if reference_vertices:
            return create_reference(
                entity_type, tuple(self.reference_vertices[i] for i in self.sub_entities(dim)[n]))
        else:
            if self.tdim == dim:
                return self
            else:
                return create_reference(
                    entity_type, tuple(self.vertices[i] for i in self.sub_entities(dim)[n]))

    def at_vertex(self, point: PointType) -> bool:
        """Check if a point is a vertex of the reference.

        Args:
            point: The point

        Returns:
            Is the point a vertex?
        """
        for v in self.vertices:
            if v == tuple(point):
                return True
        return False

    def on_edge(self, point_in: PointType) -> bool:
        """Check if a point is on an edge of the reference.

        Args:
            point_in: The point

        Returns:
            Is the point on an edge?
        """
        from .functions import VectorFunction
        point = VectorFunction(point_in)
        for e in self.edges:
            v0 = VectorFunction(self.vertices[e[0]])
            v1 = VectorFunction(self.vertices[e[1]])
            crossed = (v0 - point).cross(v1 - point).norm()
            if crossed == 0:
                return True
        return False

    def on_face(self, point_in: PointType) -> bool:
        """Check if a point is on a face of the reference.

        Args:
            point_in: The point

        Returns:
            Is the point on a face?
        """
        from .functions import VectorFunction
        point = VectorFunction(point_in)
        for f in self.faces:
            v0 = VectorFunction(self.vertices[f[0]])
            v1 = VectorFunction(self.vertices[f[1]])
            v2 = VectorFunction(self.vertices[f[2]])
            crossed = (v0 - point).cross(v1 - point)
            if crossed.dot(v2 - point):
                return True
        return False

    @abstractmethod
    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        pass

    def plot_entity_diagrams(
        self, filename: typing.Union[str, typing.List[str]],
        plot_options: typing.Dict[str, typing.Any] = {}, **kwargs: typing.Any
    ):
        """Plot diagrams showing the entity numbering of the reference."""
        from .plotting import Picture, colors

        img = Picture(**kwargs)

        if self.tdim == 1:
            offset_unit: typing.Tuple[typing.Union[
                sympy.core.expr.Expr, int], ...] = (sympy.Rational(4, 3), )
            img.add_arrow((-sympy.Rational(1, 2), 0), (-sympy.Rational(1, 3), 0))
            img.add_math((-sympy.Rational(8, 25), 0), "x", anchor="west")
        elif self.tdim == 2:
            if self.name.startswith("dual polygon"):
                offset_unit = (sympy.Rational(9, 4), 0)
                rt52 = sympy.sqrt(3) / 4
                img.add_arrow((-sympy.Rational(3, 4), -rt52), (-sympy.Rational(7, 12), -rt52))
                img.add_arrow((-sympy.Rational(3, 4), -rt52),
                              (-sympy.Rational(3, 4), sympy.Rational(1, 6) - rt52))
                img.add_math((-sympy.Rational(171, 300), -rt52), "x", anchor="west")
                img.add_math((-sympy.Rational(3, 4), sympy.Rational(9, 50) - rt52), "y",
                             anchor="south")
            else:
                offset_unit = (sympy.Rational(4, 3), 0)
                img.add_arrow((-sympy.Rational(1, 2), 0), (-sympy.Rational(1, 3), 0))
                img.add_arrow((-sympy.Rational(1, 2), 0),
                              (-sympy.Rational(1, 2), sympy.Rational(1, 6)))
                img.add_math((-sympy.Rational(8, 25), 0), "x", anchor="west")
                img.add_math((-sympy.Rational(1, 2), sympy.Rational(9, 50)), "y", anchor="south")
        elif self.tdim == 3:
            if self.name == "pyramid":
                offset_unit = (sympy.Rational(17, 12), sympy.Rational(17, 30), 0)
            elif self.name == "hexahedron":
                offset_unit = (sympy.Rational(3, 2), sympy.Rational(9, 15), 0)
            else:
                offset_unit = (1, sympy.Rational(2, 5), 0)
            img.add_arrow((-sympy.Rational(3, 8), -sympy.Rational(3, 20), 0),
                          (-sympy.Rational(5, 24), -sympy.Rational(3, 20), 0))
            img.add_arrow((-sympy.Rational(3, 8), -sympy.Rational(3, 20), 0),
                          (-sympy.Rational(3, 8), sympy.Rational(1, 60), 0))
            img.add_arrow((-sympy.Rational(3, 8), -sympy.Rational(3, 20), 0),
                          (-sympy.Rational(3, 8), -sympy.Rational(3, 20), sympy.Rational(1, 6)))
            img.add_math((-sympy.Rational(39, 200), -sympy.Rational(3, 20), 0), "x", anchor="west")
            img.add_math((-sympy.Rational(3, 8), sympy.Rational(1, 60), 0), "y",
                         anchor="south west")
            img.add_math((-sympy.Rational(3, 8), -sympy.Rational(3, 20), sympy.Rational(9, 50)),
                         "z", anchor="south")
        else:
            raise ValueError("Unsupported tdim")

        def offset(point, cd):
            return tuple((p + cd * a) / 2 for p, a in zip(point, offset_unit))

        for current_dim in range(self.tdim + 1):
            for entities in self.z_ordered_entities():
                for dim, n in entities:
                    if dim == 1:
                        start, end = [offset(self.vertices[i], current_dim) for i in self.edges[n]]
                        img.add_line(start, end)

                    if dim == current_dim:
                        ref = self.sub_entity(dim, n)
                        img.add_ncircle(offset(ref.midpoint(), current_dim), n, colors.entity(dim))

        img.save(filename, plot_options)


class Point(Reference):
    """A point."""

    def __init__(self, vertices: SetOfPointsInput = ((), )):
        """Create a point.

        Args:
            vertices: The vertices of the point.
        """
        assert len(vertices) == 1
        super().__init__(
            tdim=0,
            name="point",
            origin=vertices[0],
            axes=(),
            reference_vertices=((), ),
            vertices=vertices,
            edges=(),
            faces=(),
            volumes=(),
            sub_entity_types=["point", None, None, None],
            simplex=True, tp=True)

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Point(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        raise NotImplementedError()

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return list(zip(vars, self.vertices[0]))

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return parse_point_input(vertices[0])

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        return self.vertices[0]

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return self.vertices[0]

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        return self.reference_vertices[0]

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return sympy.Integer(0)

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.tdim == 0:
            return point in self.vertices
        raise NotImplementedError()


class Interval(Reference):
    """An interval."""

    def __init__(self, vertices: SetOfPointsInput = ((0,), (1,))):
        """Create an interval.

        Args:
            vertices: The vertices of the interval.
        """
        assert len(vertices) == 2
        super().__init__(
            tdim=1,
            name="interval",
            origin=vertices[0],
            axes=(_vsub(vertices[1], vertices[0]),),
            reference_vertices=((0,), (1,)),
            vertices=vertices,
            edges=((0, 1),),
            faces=(),
            volumes=(),
            sub_entity_types=["point", "interval", None, None],
            simplex=True, tp=True)

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Interval(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        assert self.vertices == self.reference_vertices
        return tuple((sympy.Rational(2 * i + 1, 2 * (n + 1)), ) for i in range(n))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        assert self.vertices == self.reference_vertices
        pts = tuple((sympy.Rational(i, n - 1), ) for i in range(n))
        pairs = [(i, i + 1) for i in range(n - 1)]
        return pts, pairs

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[0], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(v0 + (v1 - v0) * x[0] for v0, v1 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        p = _vsub(tuple(x), vertices[0])
        v = _vsub(vertices[1], vertices[0])
        return (_vdot(p, v) * sympy.Integer(1) / _vdot(v, v), )

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(v0 + (v1 - v0) * x[0] for v0, v1 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        p = _vsub(tuple(x), self.vertices[0])
        v = _vsub(self.vertices[1], self.vertices[0])
        return (_vdot(p, v) * sympy.Integer(1) / _vdot(v, v), )

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return 0 <= point[0] <= 1


class Triangle(Reference):
    """A triangle."""

    def __init__(self, vertices: SetOfPointsInput = ((0, 0), (1, 0), (0, 1))):
        """Create a triangle.

        Args:
            vertices: The vertices of the triangle.
        """
        assert len(vertices) == 3
        super().__init__(
            tdim=2,
            name="triangle",
            origin=vertices[0],
            axes=(_vsub(vertices[1], vertices[0]), _vsub(vertices[2], vertices[0])),
            reference_vertices=((0, 0), (1, 0), (0, 1)),
            vertices=vertices,
            edges=((1, 2), (0, 2), (0, 1)),
            faces=((0, 1, 2),),
            volumes=(),
            sub_entity_types=["point", "interval", "triangle", None],
            simplex=True)

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Triangle(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        return tuple(tuple(
            o + ((2 * i + 1) * a0 + (2 * j + 1) * a1) / 2 / (n + 1)
            for o, a0, a1 in zip(self.origin, *self.axes)
        ) for i in range(n) for j in range(n - i))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        pts = tuple(tuple(
            o + (i * a0 + j * a1) / (n - 1)
            for o, a0, a1 in zip(self.origin, *self.axes)
        ) for i in range(n) for j in range(n - i))
        pairs = []
        s = 0
        for j in range(n-1, 0, -1):
            pairs += [(i, i + 1) for i in range(s, s + j)]
            s += j + 1
        for k in range(n + 1):
            s = k
            for i in range(n, k, -1):
                if i != k + 1:
                    pairs += [(s, s + i)]
                if k != 0:
                    pairs += [(s, s + i - 1)]
                s += i
        return pts, pairs

    def z_ordered_entities_extra_dim(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities in back-to-front plotting order when using an extra dimension.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [[(1, 0), (1, 1)], [(0, 2)], [(2, 0), (1, 2), (0, 0), (0, 1)]]

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[1], sympy.Integer(0), 1 - vars[0]),
                (vars[0], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(v0 + (v1 - v0) * x[0] + (v2 - v0) * x[1] for v0, v1, v2 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert len(vertices[0]) == 2
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0]],
                            [v1[1], v2[1]]]).inv()
        return (_vdot(mat.row(0), p), _vdot(mat.row(1), p))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(v0 + (v1 - v0) * x[0] + (v2 - v0) * x[1]
                     for v0, v1, v2 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        if len(self.vertices[0]) == 2:
            p = _vsub(tuple(x), self.vertices[0])
            v1 = _vsub(self.vertices[1], self.vertices[0])
            v2 = _vsub(self.vertices[2], self.vertices[0])
            mat = sympy.Matrix([[v1[0], v2[0]],
                                [v1[1], v2[1]]]).inv()
            return (_vdot(mat.row(0), p), _vdot(mat.row(1), p))

        return tuple(
            _vdot(_vsub(tuple(x), self.origin), a) * sympy.Integer(1) / _vnorm(a) for a in self.axes
        )

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return sympy.Rational(1, 2) * self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return 0 <= point[0] and 0 <= point[1] and sum(point) <= 1


class Tetrahedron(Reference):
    """A tetrahedron."""

    def __init__(self, vertices: SetOfPointsInput = ((0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1))):
        """Create a tetrahedron.

        Args:
            vertices: The vertices of the tetrahedron.
        """
        assert len(vertices) == 4
        super().__init__(
            tdim=3,
            name="tetrahedron",
            origin=vertices[0],
            axes=(
                _vsub(vertices[1], vertices[0]),
                _vsub(vertices[2], vertices[0]),
                _vsub(vertices[3], vertices[0]),
            ),
            reference_vertices=((0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1)),
            vertices=vertices,
            edges=((2, 3), (1, 3), (1, 2), (0, 3), (0, 2), (0, 1)),
            faces=((1, 2, 3), (0, 2, 3), (0, 1, 3), (0, 1, 2)),
            volumes=((0, 1, 2, 3),),
            sub_entity_types=["point", "interval", "triangle", "tetrahedron"],
            simplex=True)

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return (self.vertices[0], self.vertices[1], self.vertices[3])

    def z_ordered_entities(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities of the cell in back-to-front plotting order.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [
            [(2, 0), (2, 1), (2, 3), (1, 0), (1, 2), (1, 4), (0, 2)],
            [(3, 0)],
            [(2, 2), (1, 1), (1, 3), (1, 5), (0, 0), (0, 1), (0, 3)]
        ]

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Tetrahedron(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        return tuple(tuple(
            o + ((2 * i + 1) * a0 + (2 * j + 1) * a1 + (2 * k + 1) * a2) / 2 / (n + 1)
            for o, a0, a1, a2 in zip(self.origin, *self.axes)
        ) for i in range(n) for j in range(n - i) for k in range(n - i - j))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[0], sympy.Integer(0), 1 - vars[1] - vars[2]),
                (vars[1], sympy.Integer(0), 1 - vars[2]),
                (vars[2], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(v0 + (v1 - v0) * x[0] + (v2 - v0) * x[1] + (v3 - v0) * x[2]
                     for v0, v1, v2, v3 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert len(vertices[0]) == 3
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])
        v3 = _vsub(vertices[3], vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return (_vdot(mat.row(0), p), _vdot(mat.row(1), p), _vdot(mat.row(2), p))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(v0 + (v1 - v0) * x[0] + (v2 - v0) * x[1] + (v3 - v0) * x[2]
                     for v0, v1, v2, v3 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        p = _vsub(tuple(x), self.vertices[0])
        v1 = _vsub(self.vertices[1], self.vertices[0])
        v2 = _vsub(self.vertices[2], self.vertices[0])
        v3 = _vsub(self.vertices[3], self.vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return (_vdot(mat.row(0), p), _vdot(mat.row(1), p), _vdot(mat.row(2), p))

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return sympy.Rational(1, 6) * self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return 0 <= point[0] and 0 <= point[1] and 0 <= point[2] and sum(point) <= 1


class Quadrilateral(Reference):
    """A quadrilateral."""

    def __init__(self, vertices: SetOfPointsInput = ((0, 0), (1, 0), (0, 1), (1, 1))):
        """Create a quadrilateral.

        Args:
            vertices: The vertices of the quadrilateral.
        """
        assert len(vertices) == 4
        super().__init__(
            tdim=2,
            name="quadrilateral",
            origin=vertices[0],
            axes=(_vsub(vertices[1], vertices[0]), _vsub(vertices[2], vertices[0])),
            reference_vertices=((0, 0), (1, 0), (0, 1), (1, 1)),
            vertices=vertices,
            edges=((0, 1), (0, 2), (1, 3), (2, 3)),
            faces=((0, 1, 2, 3),),
            volumes=(),
            sub_entity_types=["point", "interval", "quadrilateral", None],
            tp=True)

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return (self.vertices[0], self.vertices[1], self.vertices[3], self.vertices[2])

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Quadrilateral(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        return tuple(tuple(
            o + ((2 * i + 1) * a0 + (2 * j + 1) * a1) / 2 / (n + 1)
            for o, a0, a1 in zip(self.origin, *self.axes)
        ) for i in range(n + 1) for j in range(n + 1))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        pts = tuple(tuple(
            o + (i * a0 + j * a1) / (n - 1)
            for o, a0, a1 in zip(self.origin, *self.axes)
        ) for i in range(n) for j in range(n))
        pairs = []
        for i in range(n):
            for j in range(n):
                node = i * n + j
                if j != n - 1:
                    pairs += [(node, node + 1)]
                if i != n - 1:
                    pairs += [(node, node + n)]
                    if j != 0:
                        pairs += [(node, node + n - 1)]
        return pts, pairs

    def z_ordered_entities_extra_dim(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities in back-to-front plotting order when using an extra dimension.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [[(1, 3), (1, 1), (0, 2)], [(2, 0)], [(1, 2), (0, 3), (1, 0), (0, 0), (0, 1)]]

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[1], sympy.Integer(0), sympy.Integer(1)),
                (vars[0], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(
            (1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1) + x[1] * ((1 - x[0]) * v2 + x[0] * v3)
            for v0, v1, v2, v3 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert _vadd(vertices[0], vertices[3]) == _vadd(vertices[1], vertices[2])
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])

        if len(self.vertices[0]) == 2:
            mat = sympy.Matrix([[v1[0], v2[0]],
                                [v1[1], v2[1]]]).inv()
        elif len(self.vertices[0]) == 3:
            v3 = _vcross(v1, v2)
            mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                                [v1[1], v2[1], v3[1]],
                                [v1[2], v2[2], v3[2]]]).inv()
        else:
            raise RuntimeError("Cannot get inverse map.")

        return (_vdot(mat.row(0), p), _vdot(mat.row(1), p))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(
            (1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1) + x[1] * ((1 - x[0]) * v2 + x[0] * v3)
            for v0, v1, v2, v3 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        assert _vadd(
            self.vertices[0], self.vertices[3]) == _vadd(self.vertices[1], self.vertices[2])
        p = _vsub(tuple(x), self.vertices[0])
        v1 = _vsub(self.vertices[1], self.vertices[0])
        v2 = _vsub(self.vertices[2], self.vertices[0])

        if len(self.vertices[0]) == 2:
            mat = sympy.Matrix([[v1[0], v2[0]],
                                [v1[1], v2[1]]]).inv()
        elif len(self.vertices[0]) == 3:
            v3 = _vcross(v1, v2)
            mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                                [v1[1], v2[1], v3[1]],
                                [v1[2], v2[2], v3[2]]]).inv()
        else:
            raise RuntimeError("Cannot get inverse map.")

        return (_vdot(mat.row(0), p), _vdot(mat.row(1), p))

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return 0 <= point[0] <= 1 and 0 <= point[1] <= 1


class Hexahedron(Reference):
    """A hexahedron."""

    def __init__(self, vertices: SetOfPointsInput = (
        (0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0), (0, 0, 1), (1, 0, 1), (0, 1, 1), (1, 1, 1))
    ):
        """Create a hexahedron.

        Args:
            vertices: The vertices of the hexahedron.
        """
        assert len(vertices) == 8
        super().__init__(
            tdim=3,
            name="hexahedron",
            origin=vertices[0],
            axes=(
                _vsub(vertices[1], vertices[0]),
                _vsub(vertices[2], vertices[0]),
                _vsub(vertices[4], vertices[0]),
            ),
            reference_vertices=(
                (0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0),
                (0, 0, 1), (1, 0, 1), (0, 1, 1), (1, 1, 1)),
            vertices=vertices,
            edges=(
                (0, 1), (0, 2), (0, 4), (1, 3), (1, 5), (2, 3),
                (2, 6), (3, 7), (4, 5), (4, 6), (5, 7), (6, 7)),
            faces=(
                (0, 1, 2, 3), (0, 1, 4, 5), (0, 2, 4, 6),
                (1, 3, 5, 7), (2, 3, 6, 7), (4, 5, 6, 7)),
            volumes=((0, 1, 2, 3, 4, 5, 6, 7),),
            sub_entity_types=["point", "interval", "quadrilateral", "hexahedron"],
            tp=True)

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return (self.vertices[0], self.vertices[1], self.vertices[3], self.vertices[7],
                self.vertices[6], self.vertices[4])

    def z_ordered_entities(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities of the cell in back-to-front plotting order.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [
            [(2, 4), (2, 0), (1, 5), (2, 2), (1, 6), (1, 1), (0, 2)],
            [(3, 0)],
            [(2, 3), (1, 3), (1, 7), (0, 3)],
            [(2, 1), (1, 0), (1, 2), (1, 4), (0, 0), (0, 1)],
            [(2, 5), (1, 8), (1, 9), (1, 10), (1, 11), (0, 4), (0, 5), (0, 6), (0, 7)]
        ]

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Hexahedron(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        assert self.vertices == self.reference_vertices
        return tuple((
            sympy.Rational(2 * i + 1, 2 * (n + 1)),
            sympy.Rational(2 * j + 1, 2 * (n + 1)),
            sympy.Rational(2 * k + 1, 2 * (n + 1))
        ) for i in range(n + 1) for j in range(n + 1) for k in range(n + 1))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[2], sympy.Integer(0), sympy.Integer(1)),
                (vars[1], sympy.Integer(0), sympy.Integer(1)),
                (vars[0], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(
            (1 - x[2]) * ((1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1)
                          + x[1] * ((1 - x[0]) * v2 + x[0] * v3))
            + x[2] * ((1 - x[1]) * ((1 - x[0]) * v4 + x[0] * v5)
                      + x[1] * ((1 - x[0]) * v6 + x[0] * v7))
            for v0, v1, v2, v3, v4, v5, v6, v7 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert len(vertices[0]) == 3
        for a, b, c, d in self.faces:
            assert _vadd(vertices[a], vertices[d]) == _vadd(vertices[b], vertices[c])
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])
        v3 = _vsub(vertices[4], vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(
            (1 - x[2]) * ((1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1)
                          + x[1] * ((1 - x[0]) * v2 + x[0] * v3))
            + x[2] * ((1 - x[1]) * ((1 - x[0]) * v4 + x[0] * v5)
                      + x[1] * ((1 - x[0]) * v6 + x[0] * v7))
            for v0, v1, v2, v3, v4, v5, v6, v7 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        assert len(self.vertices[0]) == 3
        for a, b, c, d in self.faces:
            assert _vadd(
                self.vertices[a], self.vertices[d]) == _vadd(self.vertices[b], self.vertices[c])
        p = _vsub(tuple(x), self.vertices[0])
        v1 = _vsub(self.vertices[1], self.vertices[0])
        v2 = _vsub(self.vertices[2], self.vertices[0])
        v3 = _vsub(self.vertices[4], self.vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return 0 <= point[0] <= 1 and 0 <= point[1] <= 1 and 0 <= point[2] <= 1


class Prism(Reference):
    """A (triangular) prism."""

    def __init__(self, vertices: SetOfPointsInput = (
        (0, 0, 0), (1, 0, 0), (0, 1, 0), (0, 0, 1), (1, 0, 1), (0, 1, 1))
    ):
        """Create a prism.

        Args:
            vertices: The vertices of the prism.
        """
        assert len(vertices) == 6
        super().__init__(
            tdim=3,
            name="prism",
            origin=vertices[0],
            axes=(
                _vsub(vertices[1], vertices[0]),
                _vsub(vertices[2], vertices[0]),
                _vsub(vertices[3], vertices[0]),
            ),
            reference_vertices=(
                (0, 0, 0), (1, 0, 0), (0, 1, 0),
                (0, 0, 1), (1, 0, 1), (0, 1, 1)),
            vertices=vertices,
            edges=(
                (0, 1), (0, 2), (0, 3), (1, 2), (1, 4),
                (2, 5), (3, 4), (3, 5), (4, 5)),
            faces=(
                (0, 1, 2), (0, 1, 3, 4), (0, 2, 3, 5),
                (1, 2, 4, 5), (3, 4, 5)),
            volumes=((0, 1, 2, 3, 4, 5),),
            sub_entity_types=[
                "point", "interval",
                ["triangle", "quadrilateral", "quadrilateral", "quadrilateral", "triangle"],
                "prism"],
            tp=True)

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return (self.vertices[0], self.vertices[1], self.vertices[4], self.vertices[5],
                self.vertices[3])

    def z_ordered_entities(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities of the cell in back-to-front plotting order.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [
            [(2, 0), (2, 3), (1, 3)],
            [(2, 2), (1, 1), (1, 5), (0, 2)],
            [(3, 0)],
            [(2, 1), (1, 0), (1, 2), (1, 4), (0, 0), (0, 1)],
            [(2, 4), (1, 6), (1, 7), (1, 8), (0, 3), (0, 4), (0, 5)]
        ]

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Prism(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        assert self.vertices == self.reference_vertices
        return tuple((
            sympy.Rational(2 * i + 1, 2 * (n + 1)),
            sympy.Rational(2 * j + 1, 2 * (n + 1)),
            sympy.Rational(2 * k + 1, 2 * (n + 1))
        ) for i in range(n + 1) for j in range(n + 1 - i) for k in range(n + 1))

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[2], sympy.Integer(0), sympy.Integer(1)),
                (vars[1], sympy.Integer(0), sympy.Integer(1) - vars[0]),
                (vars[0], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(
            (1 - x[2]) * (v0 + x[0] * (v1 - v0) + x[1] * (v2 - v0))
            + x[2] * (v3 + x[0] * (v4 - v3) + x[1] * (v5 - v3))
            for v0, v1, v2, v3, v4, v5 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert len(vertices[0]) == 3
        for a, b, c, d in self.faces[1:4]:
            assert _vadd(vertices[a], vertices[d]) == _vadd(vertices[b], vertices[c])
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])
        v3 = _vsub(vertices[3], vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(
            (1 - x[2]) * (v0 + x[0] * (v1 - v0) + x[1] * (v2 - v0))
            + x[2] * (v3 + x[0] * (v4 - v3) + x[1] * (v5 - v3))
            for v0, v1, v2, v3, v4, v5 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        assert len(self.vertices[0]) == 3
        for a, b, c, d in self.faces[1:4]:
            assert _vadd(
                self.vertices[a], self.vertices[d]) == _vadd(self.vertices[b], self.vertices[c])
        p = _vsub(tuple(x), self.vertices[0])
        v1 = _vsub(self.vertices[1], self.vertices[0])
        v2 = _vsub(self.vertices[2], self.vertices[0])
        v3 = _vsub(self.vertices[3], self.vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return sympy.Rational(1, 2) * self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return (point[0] >= 0 and point[1] >= 0 and point[2] >= 0
                and point[2] <= 1 and point[0] + point[1] <= 1)


class Pyramid(Reference):
    """A (square-based) pyramid."""

    def __init__(self, vertices: SetOfPointsInput = (
        (0, 0, 0), (1, 0, 0), (0, 1, 0), (1, 1, 0), (0, 0, 1))
    ):
        """Create a pyramid.

        Args:
            vertices: The vertices of the pyramid.
        """
        assert len(vertices) == 5
        super().__init__(
            tdim=3,
            name="pyramid",
            origin=vertices[0],
            axes=(
                _vsub(vertices[1], vertices[0]),
                _vsub(vertices[2], vertices[0]),
                _vsub(vertices[4], vertices[0]),
            ),
            reference_vertices=(
                (0, 0, 0), (1, 0, 0), (0, 1, 0),
                (1, 1, 0), (0, 0, 1)),
            vertices=vertices,
            edges=(
                (0, 1), (0, 2), (0, 4), (1, 3),
                (1, 4), (2, 3), (2, 4), (3, 4)),
            faces=(
                (0, 1, 2, 3), (0, 1, 4), (0, 2, 4),
                (1, 3, 4), (2, 3, 4)),
            volumes=((0, 1, 2, 3, 4),),
            sub_entity_types=[
                "point", "interval",
                ["quadrilateral", "triangle", "triangle", "triangle", "triangle"],
                "pyramid"],
            tp=True)

    @property
    def clockwise_vertices(self) -> SetOfPoints:
        """Get list of vertices in clockwise order.

        Returns:
            A list of vertices
        """
        return (self.vertices[0], self.vertices[1], self.vertices[3], self.vertices[4])

    def z_ordered_entities(self) -> typing.List[typing.List[typing.Tuple[int, int]]]:
        """Get the subentities of the cell in back-to-front plotting order.

        Returns:
            List of lists of subentity dimensions and numbers
        """
        return [
            [(2, 0), (2, 4), (1, 5)],
            [(2, 2), (1, 1), (1, 6), (0, 2)],
            [(3, 0)],
            [(2, 3), (1, 3), (1, 7), (0, 3)],
            [(2, 1), (1, 0), (1, 2), (1, 4), (0, 0), (0, 1), (0, 4)]
        ]

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        return Pyramid(self.reference_vertices)

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        raise NotImplementedError()

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        return [(vars[0], sympy.Integer(0), 1 - vars[2]),
                (vars[1], sympy.Integer(0), 1 - vars[2]),
                (vars[2], sympy.Integer(0), sympy.Integer(1))]

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        assert self.vertices == self.reference_vertices
        return tuple(
            (1 - x[2]) * (
                (1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1)
                + x[1] * ((1 - x[0]) * v2 + x[0] * v3)
            ) + x[2] * v4
            for v0, v1, v2, v3, v4 in zip(*vertices))

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        assert self.vertices == self.reference_vertices
        vertices = parse_set_of_points_input(vertices_in)
        assert len(vertices[0]) == 3
        for a, b, c, d in self.faces[:1]:
            assert _vadd(vertices[a], vertices[d]) == _vadd(vertices[b], vertices[c])
        p = _vsub(tuple(x), vertices[0])
        v1 = _vsub(vertices[1], vertices[0])
        v2 = _vsub(vertices[2], vertices[0])
        v3 = _vsub(vertices[4], vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        return tuple(
            (1 - x[2]) * (
                (1 - x[1]) * ((1 - x[0]) * v0 + x[0] * v1)
                + x[1] * ((1 - x[0]) * v2 + x[0] * v3)
            ) + x[2] * v4
            for v0, v1, v2, v3, v4 in zip(*self.vertices))

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        assert len(self.vertices[0]) == 3
        for a, b, c, d in self.faces[:1]:
            assert _vadd(
                self.vertices[a], self.vertices[d]) == _vadd(self.vertices[b], self.vertices[c])
        p = _vsub(tuple(x), self.vertices[0])
        v1 = _vsub(self.vertices[1], self.vertices[0])
        v2 = _vsub(self.vertices[2], self.vertices[0])
        v3 = _vsub(self.vertices[4], self.vertices[0])
        mat = sympy.Matrix([[v1[0], v2[0], v3[0]],
                            [v1[1], v2[1], v3[1]],
                            [v1[2], v2[2], v3[2]]]).inv()
        return tuple(_vdot(mat.row(i), p) for i in range(mat.rows))

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        return sympy.Rational(1, 3) * self.jacobian()

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        if self.vertices != self.reference_vertices:
            raise NotImplementedError()
        return (point[0] >= 0 and point[1] >= 0 and point[2] >= 0
                and point[0] + point[2] <= 1 and point[1] + point[2] <= 1)


class DualPolygon(Reference):
    """A polygon on a barycentric dual grid."""

    reference_origin: PointType

    def __init__(
        self, number_of_triangles: int, vertices: typing.Optional[SetOfPointsInput] = None
    ):
        """Create a dual polygon.

        Args:
            number_of_triangles: The number of triangles that make up the dual polygon
            vertices: The vertices of the dual polygon.
        """
        self.number_of_triangles = number_of_triangles
        self.reference_origin = (sympy.Integer(0), sympy.Integer(0))
        reference_vertices = []
        for tri in range(number_of_triangles):
            angle = sympy.pi * 2 * tri / number_of_triangles
            next_angle = sympy.pi * 2 * (tri + 1) / number_of_triangles

            reference_vertices.append((sympy.cos(angle), sympy.sin(angle)))
            reference_vertices.append(
                ((sympy.cos(next_angle) + sympy.cos(angle)) / 2,
                 (sympy.sin(next_angle) + sympy.sin(angle)) / 2))

        origin: PointType = self.reference_origin
        if vertices is None:
            vertices = tuple(reference_vertices)
        else:
            assert len(vertices) == 1 + len(reference_vertices)
            origin = parse_point_input(vertices[0])
            vertices = vertices[1:]

        super().__init__(
            tdim=2,
            name="dual polygon",
            axes=(),
            origin=origin,
            vertices=vertices,
            reference_vertices=tuple(reference_vertices),
            edges=tuple((i, (i + 1) % (2 * number_of_triangles))
                        for i in range(2 * number_of_triangles)),
            faces=(tuple(range(2 * number_of_triangles)), ),
            volumes=(),
            sub_entity_types=["point", "interval", f"dual polygon({number_of_triangles})", None],

        )

    def contains(self, point: PointType) -> bool:
        """Check if a point is contained in the reference.

        Args:
            point: The point

        Returns:
            Is the point contained in the reference?
        """
        raise NotImplementedError()

    def integration_limits(self, vars: AxisVariablesNotSingle = t) -> IntLimits:
        """Get the limits for an integral over this reference.

        Args:
            vars: The variables to use for each direction

        Returns:
            Integration limits that can be passed into sympy.integrate
        """
        raise NotImplementedError()

    def get_map_to(self, vertices: SetOfPointsInput) -> PointType:
        """Get the map from the reference to a cell.

        Args:
            vertices: The vertices of a call

        Returns:
            The map
        """
        raise NotImplementedError()

    def get_inverse_map_to(self, vertices_in: SetOfPointsInput) -> PointType:
        """Get the inverse map from a cell to the reference.

        Args:
            vertices_in: The vertices of a cell

        Returns:
            The inverse map
        """
        raise NotImplementedError()

    def _compute_map_to_self(self) -> PointType:
        """Compute the map from the canonical reference to this reference.

        Returns:
            The map
        """
        raise NotImplementedError()

    def _compute_inverse_map_to_self(self) -> PointType:
        """Compute the inverse map from the canonical reference to this reference.

        Returns:
            The map
        """
        raise NotImplementedError()

    def volume(self) -> sympy.core.expr.Expr:
        """Calculate the volume.

        Returns:
            The volume of the cell
        """
        raise NotImplementedError()

    def default_reference(self) -> Reference:
        """Get the default reference for this cell type.

        Returns:
            The default reference
        """
        raise NotImplementedError()

    def make_lattice(self, n: int) -> SetOfPoints:
        """Make a lattice of points.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points offset from the edge of the cell
        """
        assert self.vertices == self.reference_vertices
        from .create import create_reference
        lattice: SetOfPoints = ()
        for v1, v2 in zip(self.vertices, self.vertices[1:] + (self.vertices[0], )):
            ref = create_reference("triangle", (self.origin, v1, v2))
            lattice += ref.make_lattice(n // 2)
        return lattice

    def make_lattice_with_lines(self, n: int) -> LatticeWithLines:
        """Make a lattice of points, and a list of lines connecting them.

        Args:
            n: The number of points along each edge

        Returns:
            A lattice of points including the edges of the cell
            Pairs of point numbers that make a mesh of lines across the cell
        """
        raise NotImplementedError()



"""P1-iso-P2 elements.

This element's definition appears in https://doi.org/10.1007/BF01399555
(Bercovier, Pironneau, 1979)
"""

import typing

import sympy

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..geometry import SetOfPoints
from ..piecewise_functions import PiecewiseFunction
from ..references import Reference
from ..symbols import x as x_variables


class P1IsoP2Tri(CiarletElement):
    """P1-iso-P2 finite element on a triangle."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        half = sympy.Rational(1, 2)
        zero = sympy.Integer(0)
        one = sympy.Integer(1)
        tris: typing.List[SetOfPoints] = [
            ((zero, zero), (half, zero), (zero, half)),
            ((one, zero), (half, half), (half, zero)),
            ((zero, one), (zero, half), (half, half)),
            ((zero, half), (half, half), (half, zero)),
        ]
        poly: typing.List[FunctionInput] = []
        x = x_variables[0]
        y = x_variables[1]
        c = 1 - x - y
        for pieces in [
            {0: 2 * c - 1},
            {1: 2 * x - 1},
            {2: 2 * y - 1},
            {0: 2 * x, 1: 2 * c, 3: 1 - 2 * y},
            {0: 2 * y, 2: 2 * c, 3: 1 - 2 * x},
            {1: 2 * y, 2: 2 * x, 3: 1 - 2 * c},
        ]:
            poly.append(PiecewiseFunction(
                {q: pieces[i] if i in pieces else 0 for i, q in enumerate(tris)}, 2))

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.reference_vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
        for e_n in range(3):
            entity = reference.sub_entity(1, e_n)
            dofs.append(PointEvaluation(reference, entity.midpoint(), entity=(1, e_n)))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["P1-iso-P2", "P2-iso-P1", "iso-P2 P1"]
    references = ["triangle"]
    min_order = 1
    max_order = 1
    continuity = "C0"


class P1IsoP2Quad(CiarletElement):
    """P1-iso-P2 finite element on a quadrilateral."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        half = sympy.Rational(1, 2)
        zero = sympy.Integer(0)
        one = sympy.Integer(1)
        quads: typing.List[SetOfPoints] = [
            ((zero, zero), (half, zero), (zero, half), (half, half)),
            ((half, zero), (one, zero), (half, half), (one, half)),
            ((zero, half), (half, half), (zero, one), (half, one)),
            ((half, half), (one, half), (half, one), (one, one)),
        ]
        poly: typing.List[FunctionInput] = []
        x = x_variables[0]
        y = x_variables[1]
        for pieces in [
            {0: (1 - 2 * x) * (1 - 2 * y)},
            {1: (2 * x - 1) * (1 - 2 * y)},
            {2: (1 - 2 * x) * (2 * y - 1)},
            {3: (2 * x - 1) * (2 * y - 1)},
            {0: 2 * x * (1 - 2 * y), 1: 2 * (1 - x) * (1 - 2 * y)},
            {0: 2 * (1 - 2 * x) * y, 2: 2 * (1 - 2 * x) * (1 - y)},
            {1: 2 * (2 * x - 1) * y, 3: 2 * (2 * x - 1) * (1 - y)},
            {2: 2 * x * (2 * y - 1), 3: 2 * (1 - x) * (2 * y - 1)},
            {0: 4 * x * y, 1: 4 * (1 - x) * y, 2: 4 * x * (1 - y), 3: 4 * (1 - x) * (1 - y)},
        ]:
            poly.append(PiecewiseFunction(
                {q: pieces[i] if i in pieces else 0 for i, q in enumerate(quads)}, 2))

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.reference_vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))
        for e_n in range(4):
            entity = reference.sub_entity(1, e_n)
            dofs.append(PointEvaluation(reference, entity.midpoint(), entity=(1, e_n)))
        dofs.append(PointEvaluation(reference, reference.midpoint(), entity=(2, 0)))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, 1
        )

    names = ["P1-iso-P2", "P2-iso-P1", "iso-P2 P1"]
    references = ["quadrilateral"]
    min_order = 1
    max_order = 1
    continuity = "C0"



"""Transition elements on simplices."""

import typing
from itertools import product

import sympy

from ..finite_element import CiarletElement
from ..functionals import ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import polynomial_set_1d
from ..quadrature import get_quadrature
from ..references import Reference
from ..symbols import x
from .lagrange import Lagrange


class Transition(CiarletElement):
    """Transition finite element."""

    def __init__(self, reference: Reference, order: int,
                 edge_orders: typing.Optional[typing.List[int]] = None,
                 face_orders: typing.Optional[typing.List[int]] = None,
                 variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            edge_orders: the polynomial order for each edge
            face_orders: the polynomial order for each face
            variant: The variant of the element
        """
        if reference.name == "triangle":
            assert face_orders is None
            assert edge_orders is not None
            assert len(edge_orders) == 3
        elif reference.name == "tetrahedron":
            assert face_orders is not None
            assert edge_orders is not None
            assert len(face_orders) == 4
            assert len(edge_orders) == 6

        bubble_space = Lagrange(reference, 1)

        dofs: ListOfFunctionals = []
        for v_n, v in enumerate(reference.reference_vertices):
            dofs.append(PointEvaluation(reference, v, entity=(0, v_n)))

        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_1d(reference.tdim, 1)

        for edim in range(1, 4):
            for e_n in range(reference.sub_entity_count(edim)):
                entity = reference.sub_entity(edim, e_n)
                if edim == reference.tdim:
                    entity_order = order
                elif edim == 1:
                    assert edge_orders is not None
                    entity_order = edge_orders[e_n]
                elif edim == 2:
                    assert face_orders is not None
                    entity_order = face_orders[e_n]
                else:
                    raise RuntimeError("Could not find order for this entity.")

                # DOFs
                points, _ = get_quadrature(variant, entity_order + 1)
                for ii in product(range(1, entity_order), repeat=edim):
                    if sum(ii) < entity_order:
                        pt = entity.get_point([points[j] for j in ii])
                        dofs.append(PointEvaluation(reference, pt, entity=(edim, e_n)))

                # Basis
                if entity_order > edim:
                    if edim == reference.tdim:
                        bubble = sympy.Integer(1)
                        for f in bubble_space.get_basis_functions():
                            bubble *= f
                    elif edim == reference.tdim - 1:
                        bubble = sympy.Integer(1)
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i != e_n:
                                bubble *= f
                    else:
                        assert edim == 1 and reference.tdim == 3
                        bubble = sympy.Integer(1)
                        for i, f in enumerate(bubble_space.get_basis_functions()):
                            if i in reference.edges[e_n]:
                                bubble *= f
                    space = Lagrange(entity, entity_order - edim - 1, variant=variant)
                    variables = []
                    origin = entity.vertices[0]
                    used = []
                    for p in entity.vertices[1:]:
                        i = 0
                        while p[i] == origin[i] or origin[i] == 1 or i in used:
                            i += 1
                        used.append(i)
                        variables.append(origin[i] + (p[i] - origin[i]) * x[i])
                    poly += [f.subs(x, variables) * bubble for f in space.get_basis_functions()]

        super().__init__(reference, order, poly, dofs, reference.tdim, 1)
        self.variant = variant
        self.face_orders = face_orders
        self.edge_orders = edge_orders

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant, "face_orders": self.face_orders,
                "edge_orders": self.edge_orders}

    names = ["transition"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "C0"



"""Values for Guzman-Neilan element."""

import sympy

coeffs = [
    [
        sympy.Rational(1, 6), sympy.Rational(1, 6), sympy.Rational(5, 24), sympy.Rational(-1, 24),
        sympy.Rational(5, 24), sympy.Rational(5, 24), sympy.Rational(-1, 24), sympy.Rational(5, 24),
    ],
    [
        sympy.Rational(1, 3), sympy.Rational(-2, 3), sympy.Rational(5, 12), sympy.Rational(-1, 3),
        sympy.Rational(-1, 12), sympy.Rational(-1, 3), sympy.Rational(5, 12), sympy.Rational(-5, 6),
    ],
    [
        sympy.Rational(2, 3), sympy.Rational(-1, 3), sympy.Rational(5, 6), sympy.Rational(-5, 12),
        sympy.Rational(1, 3), sympy.Rational(1, 12), sympy.Rational(1, 3), sympy.Rational(-5, 12),
    ],
]



"""Symfem: a symbolic finite element definition library."""

from .create import add_element, create_element, create_reference
from .version import version as __version__

__citation__ = "https://doi.org/10.21105/joss.03556 (Scroggs, 2021)"
__github__ = "https://github.com/mscroggs/symfem"
__git__ = "https://github.com/mscroggs/symfem.git"



"""Bogner-Fox-Schmit elements on tensor products.

This element's definition appears in http://contrails.iit.edu/reports/8569
(Bogner, Fox, Schmit, 1966)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import DerivativePointEvaluation, ListOfFunctionals, PointEvaluation
from ..functions import FunctionInput
from ..polynomials import quolynomial_set_1d
from ..references import Reference


class BognerFoxSchmit(CiarletElement):
    """Bogner-Fox-Schmit finite element."""

    def __init__(self, reference: Reference, order: int):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
        """
        assert order == 3
        dofs: ListOfFunctionals = []
        for v_n, vs in enumerate(reference.vertices):
            dofs.append(PointEvaluation(reference, vs, entity=(0, v_n)))
            for i in range(reference.tdim):
                dofs.append(DerivativePointEvaluation(
                    reference, vs, tuple(1 if i == j else 0 for j in range(reference.tdim)),
                    entity=(0, v_n)))

            if reference.tdim == 2:
                dofs.append(DerivativePointEvaluation(reference, vs, (1, 1), entity=(0, v_n),
                                                      mapping="identity"))

        poly: typing.List[FunctionInput] = []
        poly += quolynomial_set_1d(reference.tdim, order)
        super().__init__(reference, order, poly, dofs, reference.tdim, 1)

    names = ["Bogner-Fox-Schmit", "BFS"]
    references = ["quadrilateral"]
    min_order = 3
    max_order = 3
    continuity = "C1"



"""Brezzi-Douglas-Duran-Fortin elements.

This element's definition appears in https://doi.org/10.1007/BF01396752
(Brezzi, Douglas, Duran, Fortin, 1987)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput, VectorFunction
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import x
from .dpc import DPC, VectorDPC


def bddf_polyset(reference: Reference, order: int) -> typing.List[FunctionInput]:
    """Create the polynomial basis for a BDDF element.

    Args:
        reference: The reference cell
        order: The polynomial order

    Returns:
        The polynomial basis
    """
    dim = reference.tdim
    assert reference.name == "hexahedron"
    pset: typing.List[FunctionInput] = []
    pset += polynomial_set_vector(dim, dim, order)
    pset.append(VectorFunction((0, 0, x[0] ** (order + 1) * x[1])).curl())
    pset.append(VectorFunction((0, x[0] * x[2] ** (order + 1), 0)).curl())
    pset.append(VectorFunction((x[1] ** (order + 1) * x[2], 0, 0)).curl())
    for i in range(1, order + 1):
        pset.append(VectorFunction((0, 0, x[0] * x[1] ** (i + 1) * x[2] ** (order - i))).curl())
        pset.append(VectorFunction((0, x[0] ** (i + 1) * x[1] ** (order - i) * x[2], 0)).curl())
        pset.append(VectorFunction((x[0] ** (order - i) * x[1] * x[2] ** (i + 1), 0, 0)).curl())

    return pset


class BDDF(CiarletElement):
    """Brezzi-Douglas-Duran-Fortin Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly = bddf_polyset(reference, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DPC, order, {"variant": variant}),
            cells=(IntegralMoment, VectorDPC, order - 2, {"variant": variant})
        )

        self.variant = variant

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Brezzi-Douglas-Duran-Fortin", "BDDF"]
    references = ["hexahedron"]
    min_order = 1
    continuity = "H(div)"



"""Raviart-Thomas elements on simplices.

This element's definition appears in https://doi.org/10.1007/BF01396415
(Nedelec, 1980)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import IntegralMoment, ListOfFunctionals, NormalIntegralMoment
from ..functions import FunctionInput
from ..moments import make_integral_moment_dofs
from ..polynomials import Hdiv_polynomials, polynomial_set_vector
from ..references import Reference
from .lagrange import Lagrange, VectorLagrange


class RaviartThomas(CiarletElement):
    """Raviart-Thomas Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        poly += Hdiv_polynomials(reference.tdim, reference.tdim, order)

        dofs: ListOfFunctionals = make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, Lagrange, order - 1,
                    {"variant": variant}),
            cells=(IntegralMoment, VectorLagrange, order - 2, "contravariant",
                   {"variant": variant}),
        )

        super().__init__(reference, order, poly, dofs, reference.tdim, reference.tdim)
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["Raviart-Thomas", "RT", "N1div"]
    references = ["triangle", "tetrahedron"]
    min_order = 1
    continuity = "H(div)"



"""Plotting."""

import typing
from abc import ABC, abstractmethod

import sympy

from .functions import AnyFunction, VectorFunction
from .geometry import (PointType, PointTypeInput, SetOfPoints, SetOfPointsInput, parse_point_input,
                       parse_set_of_points_input)

PointOrFunction = typing.Union[PointTypeInput, AnyFunction]
SetOfPointsOrFunctions = typing.Union[
    typing.List[PointOrFunction], typing.Tuple[PointOrFunction, ...]]


def tex_font_size(n: int):
    """Convert a font size to a TeX size command.

    Args:
        n: Font size

    Returns:
        TeX size command
    """
    if n < 21:
        return "\\tiny"
    if n < 24:
        return "\\scriptsize"
    if n < 27:
        return "\\footnotesize"
    if n < 30:
        return "\\small"
    if n < 33:
        return "\\normalsize"
    if n < 36:
        return "\\large"
    if n < 39:
        return "\\Large"
    if n < 42:
        return "\\LARGE"
    if n < 45:
        return "\\huge"
    return "\\Huge"


class Colors:
    """Class storing colours used in diagrams."""

    BLACK = "#000000"
    WHITE = "#FFFFFF"
    ORANGE = "#FF8800"
    BLUE = "#44AAFF"
    GREEN = "#55FF00"
    PURPLE = "#DD2299"
    GRAY = "#AAAAAA"

    def __init__(self):
        """Initialise."""
        self._tikz = {}

    def entity(self, n: int) -> str:
        """Get the color used for an entity of a given dimension.

        Args:
            n: The dimension of the entity

        Returns:
            The color used for entities of the given dimension
        """
        if n == 0:
            return self.ORANGE
        if n == 1:
            return self.BLUE
        if n == 2:
            return self.GREEN
        if n == 3:
            return self.PURPLE
        raise ValueError(f"Unsupported dimension: {n}")

    def get_tikz_name(self, name: str) -> str:
        """Get the name of the colour to be used in Tikz.

        Args:
            name: HTML name of the color

        Returns:
            The Tikz name of the color
        """
        if name.startswith("#"):
            if name not in self._tikz:
                self._tikz[name] = f"customcolor{len(self._tikz)}"
            return self._tikz[name]
        return name

    def get_tikz_definitions(self) -> str:
        """Get the definitions of colours used in Tikz.

        Returns:
            Definitions of Tikz colors
        """
        out = ""
        for a, b in self._tikz.items():
            assert a.startswith("#")
            out += f"\\definecolor{{{b}}}{{HTML}}{{{a[1:]}}}\n"
        return out


colors = Colors()


class PictureElement(ABC):
    """An element in a picture."""

    def __init__(self):
        """Create an element."""
        pass

    @abstractmethod
    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        pass

    @abstractmethod
    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        pass

    @property
    @abstractmethod
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        pass

    def minx(self) -> sympy.core.expr.Expr:
        """Get the minimum x-coordinate.

        Returns:
            The minimum x-coordinate
        """
        return min(p[0] for p in self.points)

    def miny(self) -> sympy.core.expr.Expr:
        """Get the minimum y-coordinate.

        Returns:
            The minimum y-coordinate
        """
        return min(p[1] for p in self.points)

    def maxx(self) -> sympy.core.expr.Expr:
        """Get the maximum x-coordinate.

        Returns:
            The maximum x-coordinate
        """
        return max(p[0] for p in self.points)

    def maxy(self) -> sympy.core.expr.Expr:
        """Get the maximum y-coordinate.

        Returns:
            The maximum y-coordinate
        """
        return max(p[1] for p in self.points)


class Line(PictureElement):
    """A line."""

    def __init__(
        self, start: PointType, end: PointType, color: str,
        width: float
    ):
        """Create a line.

        Args:
            start: The start point
            end: The end point
            color: The color
            width: The width of the line
        """
        super().__init__()
        self.start = start
        self.end = end
        self.color = color
        self.width = width

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        s = map_pt(self.start)
        e = map_pt(self.end)
        return (f"<line x1='{s[0]}' y1='{s[1]}' x2='{e[0]}' y2='{e[1]}' "
                f"stroke='{self.color}' stroke-width='{self.width}' stroke-linecap='round' />\n")

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        s = map_pt(self.start)
        e = map_pt(self.end)
        return (f"\\draw[{colors.get_tikz_name(self.color)},line width={self.width * 0.2}pt,"
                f"line cap=round] ({s[0]},{s[1]}) -- ({e[0]},{e[1]});\n")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return (self.start, self.end)


class Bezier(PictureElement):
    """A Bezier curve."""

    def __init__(
        self, start: PointType, mid1: PointType, mid2: PointType, end: PointType, color: str,
        width: float
    ):
        """Create a Bezier curve.

        Args:
            start: The start point
            mid1: The first control point
            mid2: The second control point
            end: The end point
            color: The color
            width: The width of the line
        """
        super().__init__()
        self.start = start
        self.mid1 = mid1
        self.mid2 = mid2
        self.end = end
        self.color = color
        self.width = width

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        s = map_pt(self.start)
        m1 = map_pt(self.mid1)
        m2 = map_pt(self.mid2)
        e = map_pt(self.end)
        return (f"<path d='M{s[0]},{s[1]} C{m1[0]},{m1[1]}, {m2[0]},{m2[1]}, {e[0]},{e[1]}' "
                f"stroke='{self.color}' stroke-width='{self.width}' stroke-linecap='round' "
                " fill='none' />\n")

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        s = map_pt(self.start)
        m1 = map_pt(self.mid1)
        m2 = map_pt(self.mid2)
        e = map_pt(self.end)
        return (f"\\draw[{colors.get_tikz_name(self.color)},line width={self.width * 0.2}pt,"
                f"line cap=round] ({s[0]},{s[1]}) .. controls ({m1[0]},{m1[1]}) "
                f"and ({m2[0]},{m2[1]}) .. ({e[0]},{e[1]});\n")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return (self.start, self.end)


class Arrow(PictureElement):
    """An arrow."""

    def __init__(
        self, start: PointType, end: PointType, color: str,
        width: float
    ):
        """Create an arrow.

        Args:
            start: The start point
            end: The end point
            color: The color
            width: The width of the line
        """
        super().__init__()
        self.start = start
        self.end = end
        self.color = color
        self.width = width

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        out = ""

        s = map_pt(self.start)
        e = map_pt(self.end)
        out += (f"<line x1='{s[0]}' y1='{s[1]}' x2='{e[0]}' y2='{e[1]}' "
                f"stroke='{self.color}' stroke-width='{self.width}' stroke-linecap='round' />\n")

        ve = VectorFunction(self.end)
        vs = VectorFunction(self.start)
        vdirection = ve - vs
        vdirection /= vdirection.norm()
        vdirection /= 30
        perp = VectorFunction((-vdirection[1], vdirection[0]))
        perp /= sympy.Rational(5, 2)
        for adir in [vdirection + perp, vdirection - perp]:
            pt = ve - adir * sympy.S(self.width) / 4
            pt_s = pt.as_sympy()
            assert isinstance(pt_s, tuple)
            m = map_pt(pt_s)
            out += (f"<line x1='{m[0]}' y1='{m[1]}' x2='{e[0]}' y2='{e[1]}' "
                    f"stroke='{self.color}' stroke-width='{self.width}' stroke-linecap='round'"
                    " />\n")
        return out

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        s = map_pt(self.start)
        e = map_pt(self.end)
        return (f"\\draw[-stealth,{colors.get_tikz_name(self.color)},"
                f"line width={self.width * 0.2}pt,line cap=round] "
                f"({s[0]},{s[1]}) -- ({e[0]},{e[1]});\n")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return (self.start, self.end)


class NCircle(PictureElement):
    """A circle containing a number."""

    def __init__(
        self, centre: PointType, number: int, color: str, text_color: str, fill_color: str,
        radius: float, font_size: typing.Union[int, None], width: float, font: str
    ):
        """Create a circle containing a number.

        Args:
            centre: The centre of the circle
            number: The number to print inside the circle
            color: The colour of the circle's line
            text_color: The text colour
            fill_color: The colour to fill the circle
            radius: The radius of the circle
            font_size: The font size
            width: The width of the line
            font: The font to use for the number
        """
        self.centre = centre
        self.number = number
        self.color = color
        self.text_color = text_color
        self.fill_color = fill_color
        self.radius = radius
        self.font = font
        if font_size is None:
            if number < 10:
                self.font_size = 25
            elif number < 100:
                self.font_size = 20
            else:
                self.font_size = 12
        else:
            self.font_size = font_size
        self.width = width

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        c = map_pt(self.centre)

        return (f"<circle cx='{c[0]}' cy='{c[1]}' r='{self.radius}' "
                f"stroke='{self.color}' stroke-width='{self.width}' fill='{self.fill_color}' />\n"
                f"<text x='{c[0]}' y='{c[1]}' fill='{self.text_color}' font-size='{self.font_size}'"
                f" style=\"text-anchor:middle;dominant-baseline:middle;font-family:{self.font}\">"
                f"{self.number}</text>\n")

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        c = map_pt(self.centre)
        return (f"\\draw[{colors.get_tikz_name(self.color)},line width={self.width * 0.2}pt,"
                f"fill={colors.get_tikz_name(self.fill_color)}] "
                f"({c[0]},{c[1]}) circle ({self.radius * 0.2}pt);\n"
                f"\\node[{colors.get_tikz_name(self.text_color)},anchor=center] "
                f"at ({c[0]},{c[1]}) {{{tex_font_size(self.font_size)} {self.number}}};")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return (self.centre, )


class Fill(PictureElement):
    """A filled polygon."""

    def __init__(self, vertices: SetOfPoints, color: str, opacity: float):
        """Create a filled polygon.

        Args:
            vertices: The vertices of the polygon in clockwise or anticlockwise order
            color: The colour to fill the polygon
            opacity: The opacity
        """
        self.vertices = vertices
        self.color = color
        self.opacity = opacity

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        pts = [map_pt(p) for p in self.vertices]
        ptstring = " ".join(f"{p[0]},{p[1]}" for p in pts)
        return f"<polygon points='{ptstring}' fill='{self.color}' opacity='{self.opacity}' />"

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        vs = [map_pt(v) for v in self.vertices]
        return (f"\\fill[{colors.get_tikz_name(self.color)},opacity={self.opacity}]"
                " " + " -- ".join([f"({v[0]},{v[1]})" for v in vs]) + " -- cycle;")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return self.vertices


class Math(PictureElement):
    """A mathematical symbol."""

    def __init__(self, point: PointType, math: str, color: str, font_size: int,
                 anchor: str):
        """Create a filled polygon.

        Args:
            point: The point to put the math
            math: The math
            color: The color of the math
            font_size: The font size
            anchor: The point on the equation to anchor to
        """
        self.point = parse_point_input(point)
        self.math = math
        self.color = color
        self.font_size = font_size
        assert anchor in [
            "center", "north", "south", "east", "west",
            "north east", "north west", "south east", "south west"]
        self.anchor = anchor

    def as_svg(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return SVG format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            An SVG string
        """
        if self.anchor == "center":
            anchor = "text-anchor:middle;dominant-baseline:middle"
        else:
            anchor = ""
            if self.anchor.startswith("north"):
                anchor += "dominant-baseline:top"
            elif self.anchor.startswith("south"):
                anchor += "dominant-baseline:bottom"
            else:
                anchor += "dominant-baseline:middle"
            anchor += ";"
            if self.anchor.endswith("east"):
                anchor += "text-anchor:right"
            elif self.anchor.endswith("west"):
                anchor += "text-anchor:left"
            else:
                anchor += "text-anchor:middle"

        c = map_pt(self.point)

        return (f"<text x='{c[0]}' y='{c[1]}' fill='{self.color}' font-size='{self.font_size}'"
                f" style=\"{anchor};font-family:'CMU Serif',serif;font-style:italic\">"
                f"{self.math}</text>\n")

    def as_tikz(
        self, map_pt: typing.Callable[[PointType], typing.Tuple[float, float]]
    ) -> str:
        """Return Tikz format.

        Args:
            map_pt: A function that adjust the origin and scales the picture

        Returns:
            A Tikz string
        """
        p = map_pt(self.point)
        return (f"\\node[{colors.get_tikz_name(self.color)},anchor={self.anchor}] "
                f"at ({p[0]},{p[1]}) {{{tex_font_size(self.font_size)}${self.math}$}};")

    @property
    def points(self) -> SetOfPoints:
        """Get set of points used by this element.

        Returns:
            A set of points
        """
        return (self.point, )


class Picture:
    """A picture."""

    axes_3d: SetOfPoints

    def __init__(
        self, padding: sympy.core.expr.Expr = sympy.Integer(25), scale: typing.Optional[int] = None,
        width: typing.Optional[int] = None, height: typing.Optional[int] = None,
        axes_3d: typing.Optional[SetOfPointsInput] = None,
        dof_arrow_size: typing.Union[int, sympy.core.expr.Expr] = 1,
        title: typing.Optional[str] = None, desc: typing.Optional[str] = None,
        svg_metadata: typing.Optional[str] = None, tex_comment: typing.Optional[str] = None
    ):
        """Create a picture.

        Args:
            padding: The padding between the objects and the edge of the picture
            scale: The amount to scale the coordinates by
            width: The width of the picture
            height: The height of the picture
            axes_3d: The axes to use when drawing a 3D object
            dof_arrow_size: The relative length of the DOF arrows
            title: The title of the picture
            desc: A description of the picture
            svg_metadata: Metadata to put at the start of the SVG file
            tex_comment: Comment to put at the start of the TeX file
        """
        self._default_axes: SetOfPoints = (
            (sympy.Integer(1), -sympy.Rational(2, 25)),
            (sympy.Rational(1, 2), sympy.Rational(1, 5)),
            (sympy.Integer(0), sympy.Integer(1)))
        self.elements: typing.List[PictureElement] = []
        self.padding = padding
        self.scale = scale
        self.height = height
        if isinstance(dof_arrow_size, int):
            self.dof_arrow_size = sympy.Integer(dof_arrow_size)
        else:
            assert isinstance(dof_arrow_size, sympy.core.expr.Expr)
            self.dof_arrow_size = dof_arrow_size
        self.width = width
        self.title = title
        self.desc = desc
        if svg_metadata is None:
            import symfem
            self.svg_metadata = (
                "<!--\n"
                "This diagram was created using Symfem\n"
                f"{symfem.__github__}\n"
                f"{symfem.__citation__}\n"
                "-->\n")
        else:
            assert isinstance(svg_metadata, str)
            self.svg_metadata = svg_metadata
        if tex_comment is None:
            import symfem
            self.tex_comment = (
                "% This diagram was created using Symfem\n"
                f"% {symfem.__github__}\n"
                f"% {symfem.__citation__}\n\n")
        else:
            assert isinstance(tex_comment, str)
            self.tex_comment = tex_comment

        if axes_3d is None:
            self.axes_3d = self._default_axes
        else:
            self.axes_3d = parse_set_of_points_input(axes_3d)

    def z(self, p_in: PointOrFunction) -> sympy.core.expr.Expr:
        """Get the into/out-of-the-page component of a point.

        Args:
            p_in: The point

        Returns:
            The into/out-of-the-page component of the point
        """
        p = self.parse_point(p_in)
        if len(p) == 3:
            assert self.axes_3d == self._default_axes
            return p[0] - 2 * p[1]
        return sympy.Integer(0)

    def to_2d(self, p: PointType) -> PointType:
        """Map a point to 2D.

        Args:
            p: The point

        Return:
            The projection of the point into 2 dimensions
        """
        zero = sympy.Integer(0)
        if len(p) == 0:
            return (zero, zero)
        if len(p) == 1:
            return (p[0], zero)
        if len(p) == 2:
            return p
        if len(p) == 3:
            return (
                self.axes_3d[0][0] * p[0] + self.axes_3d[1][0] * p[1] + self.axes_3d[2][0] * p[2],
                self.axes_3d[0][1] * p[0] + self.axes_3d[1][1] * p[1] + self.axes_3d[2][1] * p[2])
        raise ValueError(f"Unsupported gdim: {len(p)}")

    def parse_point(self, p: PointOrFunction) -> PointType:
        """Parse an input point.

        Args:
            p: a point or a function

        Returns:
            The point as a tuple of Sympy expressions
        """
        if isinstance(p, AnyFunction):
            p_s = p.as_sympy()
            assert isinstance(p_s, tuple)
            p = p_s
        assert isinstance(p, (tuple, list))
        return self.to_2d(parse_point_input(p))

    def add_line(
        self, start: PointOrFunction, end: PointOrFunction, color: str = colors.BLACK,
        width: float = 4.0
    ):
        """Add a line to the picture.

        Args:
            start: The start point of the line
            end: The end point of the line
            color: The color of the line
            width: The width of the line
        """
        self.elements.append(Line(self.parse_point(start), self.parse_point(end), color, width))

    def add_bezier(
        self, start: PointOrFunction, mid1: PointOrFunction, mid2: PointOrFunction,
        end: PointOrFunction, color: str = colors.BLACK, width: float = 4.0
    ):
        """Add a Bezier curve to the picture.

        Args:
            start: The start point of the Bezier curve
            mid1: The first control point
            mid2: The second control point
            end: The end point of the Bezier curve
            color: The color of the Bezier curve
            width: The width of the Bezier curve
        """
        self.elements.append(Bezier(
            self.parse_point(start), self.parse_point(mid1), self.parse_point(mid2),
            self.parse_point(end), color, width))

    def add_arrow(
        self, start: PointOrFunction, end: PointOrFunction, color: str = colors.BLACK,
        width: float = 4.0
    ):
        """Add an arrow to the picture.

        Args:
            start: The start point of the arrow
            end: The end point of the arrow
            color: The color of the arrow
            width: The width of the arrow
        """
        self.elements.append(Arrow(self.parse_point(start), self.parse_point(end), color, width))

    def add_dof_marker(
        self, point: PointOrFunction, number: int, color: str, bold: bool = True
    ):
        """Add a DOF marker.

        Args:
            point: The point
            number: The number to put in the marker
            color: The color of the marker
            bold: Should the marker be bold?
        """
        if bold:
            self.add_ncircle(point, number, colors.BLACK, colors.BLACK, color)
        else:
            self.add_ncircle(point, number, color, color, colors.WHITE)

    def add_dof_arrow(
        self, point: PointOrFunction, direction: PointOrFunction, number: int,
        color: str = colors.PURPLE, shifted: bool = False, bold: bool = True
    ):
        """Add a DOF arrow.

        Args:
            point: The point
            direction: The direction of the arrow
            number: The number to put in the marker
            color: The color of the marker
            shifted: Should the marker be shifted?
            bold: Should the marker be bold?
        """
        vdirection = VectorFunction(self.parse_point(direction))
        vdirection /= 8 * vdirection.norm()
        vdirection *= self.dof_arrow_size
        start = VectorFunction(self.parse_point(point))
        if shifted:
            start += vdirection / 3
        self.add_arrow(start, start + vdirection, color)
        self.add_dof_marker(start, number, color, bold)

    def add_ncircle(
        self, centre: PointOrFunction, number: int, color: str = "red",
        text_color: str = colors.BLACK, fill_color: str = colors.WHITE, radius: float = 20.0,
        font_size: typing.Optional[int] = None, width: float = 4.0,
        font: str = "'Varela Round',sans-serif"
    ):
        """Add a numbered circle to the picture.

        Args:
            centre: The centre points
            number: The number in the circle
            color: The color of the outline
            text_color: The color of the test
            fill_color: The colour of the background fill
            radius: The radius of the circle
            font_size: The font size
            width: The width of the line
            font: The font
        """
        self.elements.append(NCircle(
            self.parse_point(centre), number, color, text_color, fill_color, radius, font_size,
            width, font))

    def add_math(self, point: PointTypeInput, math: str, color: str = colors.BLACK,
                 font_size: int = 35, anchor="center"):
        """Create mathematical symbol.

        Args:
            point: The point to put the math
            math: The math
            color: The color of the math
            font_size: The font size
            anchor: The point on the equation to anchor to
        """
        self.elements.append(Math(self.parse_point(point), math, color, font_size, anchor))

    def add_fill(
        self, vertices: SetOfPointsOrFunctions, color: str = "red", opacity: float = 1.0
    ):
        """Add a filled polygon to the picture.

        Args:
            vertices: The vertices of the polygon
            color: The color of the polygon
            opacity: The opacity of the polygon
        """
        self.elements.append(Fill(tuple(self.parse_point(p) for p in vertices), color, opacity))

    def compute_scale(self, unit: str = "px", reverse_y: bool = True) -> typing.Tuple[
        sympy.core.expr.Expr, sympy.core.expr.Expr, sympy.core.expr.Expr,
        typing.Callable[[PointType], typing.Tuple[float, float]]
    ]:
        """Compute the scale and size of the picture.

        Args:
            unit: The unit to use. Accepted values: px, cm, mm
            reverse_y: Should the y-axis be reversed?

        Returns:
            The scale, height, and width of the image, and a mapping function
        """
        minx = min(i.minx() for i in self.elements)
        miny = min(i.miny() for i in self.elements)
        maxx = max(i.maxx() for i in self.elements)
        maxy = max(i.maxy() for i in self.elements)

        if self.scale is not None:
            assert self.width is None
            assert self.height is None
            scale = sympy.Integer(self.scale)
        elif self.width is not None:
            assert self.height is None
            scale = sympy.Integer(self.width) / (maxx - minx)
        elif self.height is not None:
            scale = sympy.Integer(self.height) / (maxy - miny)
        else:
            scale = sympy.Integer(450)

        if unit == "mm":
            scale /= 20
        elif unit == "cm":
            scale /= 200
        elif unit == "px":
            pass
        else:
            raise ValueError(f"Unknown unit: {unit}")

        width = 2 * self.padding + (maxx - minx) * scale
        height = 2 * self.padding + (maxy - miny) * scale

        if reverse_y:
            def map_pt(pt: PointType) -> typing.Tuple[float, float]:
                """Map a point."""
                return (
                    float(self.padding + (pt[0] - minx) * scale),
                    float(height - self.padding - (pt[1] - miny) * scale))
        else:
            def map_pt(pt: PointType) -> typing.Tuple[float, float]:
                """Map a point."""
                return (
                    float(self.padding + (pt[0] - minx) * scale),
                    float(self.padding + (pt[1] - miny) * scale))

        return scale, height, width, map_pt

    def as_svg(self, filename: typing.Optional[str] = None) -> str:
        """Convert to an SVG.

        Args:
            filename: The file name

        Returns:
            The image as an SVG string
        """
        scale, height, width, map_pt = self.compute_scale("px")

        assert filename is None or filename.endswith(".svg")
        img = (f"<svg width='{float(width)}' height='{float(height)}'"
               " xmlns='http://www.w3.org/2000/svg' xmlns:xlink='http://www.w3.org/1999/xlink'>\n")
        if self.title is not None:
            img += f"<title>{self.title}</title>\n"
        if self.desc is not None:
            img += f"<desc>{self.desc}</desc>\n"
        if self.svg_metadata is not None:
            img += self.svg_metadata + "\n"

        for e in self.elements:
            img += e.as_svg(map_pt)

        img += "</svg>"

        if filename is not None:
            with open(filename, "w") as f:
                f.write(img)

        return img

    def as_png(self, filename: str, png_scale: typing.Optional[float] = None,
               png_width: typing.Optional[int] = None, png_height: typing.Optional[int] = None):
        """Convert to a PNG.

        Args:
            filename: The file name
            png_scale: The scale of the png
            png_width: The width of the png
            png_height: The height of the png
        """
        try:
            from cairosvg import svg2png
        except ImportError:
            raise ImportError("CairoSVG is needed for plotting PNGs"
                              " (pip install CairoSVG)")

        if png_scale is not None:
            assert png_width is None
            assert png_height is None
        elif png_width is not None:
            assert png_height is None
            png_scale = png_width / float(self.compute_scale("px")[2])
        elif png_height is not None:
            png_scale = png_height / float(self.compute_scale("px")[1])
        else:
            png_scale = 1.0

        assert isinstance(png_scale, float)
        assert filename.endswith(".png")
        svg2png(bytestring=self.as_svg(), write_to=filename, scale=png_scale)

    def as_tikz(self, filename: typing.Optional[str] = None) -> str:
        """Convert to tikz.

        Args:
            filename: The file name

        Returns:
            The image as a Tikz string
        """
        scale, height, width, map_pt = self.compute_scale("cm", False)
        tikz = self.tex_comment
        tikz += "\\begin{tikzpicture}[x=1cm,y=1cm]\n"

        inner_tikz = ""
        for e in self.elements:
            inner_tikz += e.as_tikz(map_pt)

        tikz += colors.get_tikz_definitions() + inner_tikz

        tikz += "\\end{tikzpicture}\n"

        if filename is not None:
            with open(filename, "w") as f:
                f.write(tikz)

        return tikz

    def save(self, filename: typing.Union[str, typing.List[str]],
             plot_options: typing.Dict[str, typing.Any] = {}):
        """Save the picture as a file.

        Args:
            filename: The file name
            plot_options: The plotting options
        """
        if isinstance(filename, list):
            for fn in filename:
                self.save(fn, plot_options)
                return

        assert isinstance(filename, str)
        if filename.endswith(".svg"):
            self.as_svg(filename, **plot_options)
        elif filename.endswith(".png"):
            self.as_png(filename, **plot_options)
        elif filename.endswith(".tex"):
            self.as_tikz(filename, **plot_options)
        else:
            if "." in filename:
                ext = "." + filename.split(".")[-1]
                raise ValueError(f"Unknown file extension: {ext}")
            else:
                raise ValueError(f"Unknown file extension: {filename}")



"""Trimmed serendipity elements on tensor products.

These elements' definitions appear in https://doi.org/10.1137/16M1073352
(Cockburn, Fu, 2017) and https://doi.org/10.1090/mcom/3354
(Gilette, Kloefkorn, 2018)
"""

import typing

from ..finite_element import CiarletElement
from ..functionals import (IntegralAgainst, IntegralMoment, ListOfFunctionals, NormalIntegralMoment,
                           TangentIntegralMoment)
from ..functions import FunctionInput, ScalarFunction, VectorFunction
from ..moments import make_integral_moment_dofs
from ..polynomials import polynomial_set_vector
from ..references import Reference
from ..symbols import t, x
from .dpc import DPC, VectorDPC


class TrimmedSerendipityHcurl(CiarletElement):
    """Trimmed serendipity Hcurl finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        if reference.tdim == 2:
            poly += [
                (x[0] ** j * x[1] ** (order - j), -x[0] ** (j + 1) * x[1] ** (order - 1 - j))
                for j in range(order)]
            if order == 1:
                poly += [(x[1], x[0])]
            else:
                poly += [(x[1] ** order, order * x[0] * x[1] ** (order - 1)),
                         (order * x[0] ** (order - 1) * x[1], x[0] ** order)]
        else:
            for i in range(order):
                for j in range(order - i):
                    for dim in range(3):
                        if i == 0 or dim != 0:
                            p = VectorFunction(tuple(x)).cross(VectorFunction([
                                x[0] ** i * x[1] ** j * x[2] ** (order - 1 - i - j)
                                if d == dim else 0 for d in range(3)]))
                            poly.append(p)

            if order == 1:
                poly += [ScalarFunction(x[0] * x[1] * x[2] ** order).grad(3)]
            else:
                poly += [ScalarFunction(x[0] * x[1] * x[2] ** order).grad(3),
                         ScalarFunction(x[0] * x[1] ** order * x[2]).grad(3),
                         ScalarFunction(x[0] ** order * x[1] * x[2]).grad(3)]
            for i in range(order + 1):
                poly.append(ScalarFunction(x[0] * x[1] ** i * x[2] ** (order - i)).grad(3))
                if i != 1:
                    poly.append(
                        ScalarFunction(x[1] * x[0] ** i * x[2] ** (order - i)).grad(3))
                    if order - i != 1:
                        poly.append(
                            ScalarFunction(x[2] * x[0] ** i * x[1] ** (order - i)).grad(3))
            for i in range(order):
                p = x[0] * x[1] ** i * x[2] ** (order - 1 - i)
                poly.append((0, -x[2] * p, x[1] * p))
                p = x[1] * x[0] ** i * x[2] ** (order - 1 - i)
                poly.append((-x[2] * p, 0, x[0] * p))
                if order > 1:
                    p = x[2] * x[0] ** i * x[1] ** (order - 1 - i)
                    poly.append((-x[1] * p, x[0] * p, 0))

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            edges=(TangentIntegralMoment, DPC, order - 1, {"variant": variant}),
            faces=(IntegralMoment, VectorDPC, order - 3, "contravariant",
                   {"variant": variant}),
            volumes=(IntegralMoment, VectorDPC, order - 5, "contravariant",
                     {"variant": variant}),
        )
        if order >= 2:
            for f_n in range(reference.sub_entity_count(2)):
                face = reference.sub_entity(2, f_n)
                for i in range(order):
                    f = ScalarFunction(x[0] ** (order - 1 - i) * x[1] ** i).grad(2)
                    f2 = VectorFunction((f[1], -f[0])).subs(x, tuple(t))
                    dofs.append(IntegralAgainst(
                        reference, face, f2, entity=(2, f_n), mapping="contravariant"))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, reference.tdim
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["trimmed serendipity Hcurl", "TScurl"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(curl)"


class TrimmedSerendipityHdiv(CiarletElement):
    """Trimmed serendipity Hdiv finite element."""

    def __init__(self, reference: Reference, order: int, variant: str = "equispaced"):
        """Create the element.

        Args:
            reference: The reference element
            order: The polynomial order
            variant: The variant of the element
        """
        poly: typing.List[FunctionInput] = []
        poly += polynomial_set_vector(reference.tdim, reference.tdim, order - 1)
        if reference.tdim == 2:
            poly += [
                (x[0] ** (j + 1) * x[1] ** (order - 1 - j), x[0] ** j * x[1] ** (order - j))
                for j in range(order)]
            if order == 1:
                poly += [(x[0], -x[1])]
            else:
                poly += [(order * x[0] * x[1] ** (order - 1), -x[1] ** order),
                         (-x[0] ** order, order * x[0] ** (order - 1) * x[1])]
        else:
            for i in range(order):
                for j in range(order - i):
                    p = x[0] ** i * x[1] ** j * x[2] ** (order - 1 - i - j)
                    poly.append((x[0] * p, x[1] * p, x[2] * p))
            for i in range(order):
                p = x[0] * x[1] ** i * x[2] ** (order - 1 - i)
                poly.append(
                    VectorFunction((0, -x[2] * p, x[1] * p)).curl())
                p = x[1] * x[0] ** i * x[2] ** (order - 1 - i)
                poly.append(
                    VectorFunction((-x[2] * p, 0, x[0] * p)).curl())
                if order > 1:
                    p = x[2] * x[0] ** i * x[1] ** (order - 1 - i)
                    poly.append(
                        VectorFunction((-x[1] * p, x[0] * p, 0)).curl())

        dofs: ListOfFunctionals = []
        dofs += make_integral_moment_dofs(
            reference,
            facets=(NormalIntegralMoment, DPC, order - 1, {"variant": variant}),
            cells=(IntegralMoment, VectorDPC, order - 3, "covariant",
                   {"variant": variant}),
        )
        if order >= 2:
            if reference.tdim == 2:
                fs = [ScalarFunction(x[0] ** (order - 1 - i) * x[1] ** i).grad(2)
                      for i in range(order)]
            else:
                fs = [ScalarFunction(x[0] ** (order - 1 - i - j) * x[1] ** i * x[2] ** j).grad(3)
                      for i in range(order) for j in range(order - i)]
            for f in fs:
                f2 = f.subs(x, tuple(t))
                dofs.append(IntegralAgainst(reference, reference, f2, entity=(reference.tdim, 0),
                                            mapping="covariant"))

        super().__init__(
            reference, order, poly, dofs, reference.tdim, reference.tdim
        )
        self.variant = variant

    def init_kwargs(self) -> typing.Dict[str, typing.Any]:
        """Return the kwargs used to create this element.

        Returns:
            Keyword argument dictionary
        """
        return {"variant": self.variant}

    names = ["trimmed serendipity Hdiv", "TSdiv"]
    references = ["quadrilateral", "hexahedron"]
    min_order = 1
    continuity = "H(div)"
